#include <pwd.h>
#include <iconv.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <pthread.h>
#include <locale.h>
#include <langinfo.h>
#include "config.h"

#define CONFIGFILE	".smbnetfs"

char		*login		= "guest";
char		*local_charset	= "UTF-8";
char		*samba_charset	= "UTF-8";
char		*config		= NULL;
FILE		*stdlog		= NULL;
pthread_mutex_t	m_config	= PTHREAD_MUTEX_INITIALIZER;
struct option	opt;


char* alloc_string(char *value){
    char *ptr = NULL;

    if ((value != NULL) &&
	((ptr = malloc(strlen(value) + 1)) != NULL)) strcpy(ptr, value);
    return ptr;
}

int get_number(char *value){
    int val = 0;

    while(*value){
	if ((*value < '0') || (*value > '9')) return -1;
	val = val * 10 + (*value - '0');
	value++;
    }
    return val;
}

int get_boolean(char *value){
    if ((strcasecmp(value, "true") == 0) || (strcasecmp(value, "yes") == 0))
	return 1;
    if ((strcasecmp(value, "false") == 0) || (strcasecmp(value, "no") == 0))
	return 0;
    return -1;
}

void create_defaults(){
    register struct passwd	*pw;
    int				len;
    char			*tmp;

    check_option(&opt);
    setlocale(LC_ALL, "");
    if ((tmp = alloc_string(nl_langinfo(CODESET))) != NULL)
	local_charset = tmp;

    if ((pw = getpwuid(getuid())) != NULL)
	if ((tmp = alloc_string(pw->pw_name)) != NULL)
	    login = tmp;

    if ((tmp = getenv("HOME")) != NULL){
	len = strlen(tmp);
	if ((config = malloc(len + strlen(CONFIGFILE) + 2)) != NULL){
	    strcpy(config, tmp);
	    config[len] = '/';
	    strcpy(config + len + 1, CONFIGFILE);
	}
    }
    check_option(&opt);
}

int set_option(struct option *opt, char *name, char *value){
    if ((strcasecmp(name, "debug") == 0) &&
	((opt->debug = get_number(value)) > 0)) return 1;

    if ((strcasecmp(name, "smb_descr_max") == 0) &&
	((opt->smb_descr_max = get_number(value)) > 0)) return 1;

    if ((strcasecmp(name, "refresh_time") == 0) &&
	((opt->refresh_time = get_number(value)) > 0)) return 1;

    if ((strcasecmp(name, "dead_time") == 0) &&
	((opt->dead_time = get_number(value)) > 0)) return 1;

    if ((strcasecmp(name, "block_size") == 0) &&
	((opt->block_size = get_number(value)) > 0)) return 1;

    if ((strcasecmp(name, "quiet_flag") == 0) &&
	((opt->quiet_flag = get_boolean(value)) >= 0)) return 1;

    if ((strcasecmp(name, "query_browsers") == 0) &&
	((opt->query_browsers = get_boolean(value)) >= 0)) return 1;

    if ((strcasecmp(name, "username") == 0) &&
	(opt->username = alloc_string(value))) return 1;

    if ((strcasecmp(name, "password") == 0) &&
	(opt->password = alloc_string(value))) return 1;

    if ((strcasecmp(name, "local_charset") == 0) &&
	(opt->local_charset = alloc_string(value))) return 1;

    if ((strcasecmp(name, "samba_charset") == 0) &&
	(opt->samba_charset = alloc_string(value))) return 1;

    if ((strcasecmp(name, "log_file") == 0) &&
	(opt->log_file = alloc_string(value))) return 1;

    return 0;
}

void init_option(struct option *opt){
    opt->debug = 0;
    opt->smb_descr_max = 15;
    opt->refresh_time = 300;
    opt->dead_time = 3 * opt->refresh_time;
    opt->block_size = 32768;
    opt->quiet_flag = 1;
    opt->query_browsers = 1;
    opt->username = NULL;
    opt->password = NULL;
    opt->local_charset = NULL;
    opt->samba_charset = NULL;
    opt->log_file = NULL;
}

void destroy_option(struct option *opt){
    if (opt->username != NULL) free(opt->username);
    if (opt->password != NULL) free(opt->password);
    if (opt->local_charset != NULL) free(opt->local_charset);
    if (opt->samba_charset != NULL) free(opt->samba_charset);
    if (opt->log_file != NULL) free(opt->log_file);
}

void check_option(struct option *opt){
    if ((opt->debug < 0) || (opt->debug > 9)) opt->debug = 0;
    if (opt->smb_descr_max < 3) opt->smb_descr_max = 15;
    if (opt->refresh_time < 0) opt->refresh_time = 300;
    if (opt->dead_time < opt->refresh_time)
	opt->dead_time = 3 * opt->refresh_time;
    if (opt->block_size <= 0) opt->block_size = 32768;
    opt->block_size -= (opt->block_size % 4096);
    if (opt->block_size < 4096) opt->block_size = 4096;
    if ((opt->quiet_flag < 0) || (opt->quiet_flag > 1)) opt->quiet_flag = 1;
    if ((opt->query_browsers < 0) || (opt->query_browsers > 1))
	opt->query_browsers = 1;
    
    if (opt->username == NULL) opt->username = alloc_string(login);
    if (opt->password == NULL) opt->password = alloc_string("");
    /* opt->local_charset */
    /* opt->samba_charset */
    /* opt->log_file */
}

int option(int opts){
    pthread_mutex_lock(&m_config);
    switch(opts){
	case OPT_DEBUG:		 opts = opt.debug; break;
	case OPT_SMB_MAX_DESCR:	 opts = opt.smb_descr_max; break;
	case OPT_REFRESH_TIME:	 opts = opt.refresh_time; break;
	case OPT_DEAD_TIME:	 opts = opt.dead_time; break;
	case OPT_BLOCK_SIZE:	 opts = opt.block_size; break;
	case OPT_QUIET_FLAG:	 opts = opt.quiet_flag; break;
	case OPT_QUERY_BROWSERS: opts = opt.query_browsers; break;
	default:		 opts = -1;
    }
    pthread_mutex_unlock(&m_config);
    return opts;
}

void set_all_option(struct option *new_opt){
    pthread_mutex_lock(&m_config);
    destroy_option(&opt);
    memcpy(&opt, new_opt, sizeof(struct option));
    pthread_mutex_unlock(&m_config);
}

iconv_t local2samba(){
    char	*local = local_charset;
    char	*samba = samba_charset;
    iconv_t	cd;
    
    pthread_mutex_lock(&m_config);
    if (opt.local_charset != NULL) local = opt.local_charset;
    if (opt.samba_charset != NULL) samba = opt.samba_charset;
    cd = iconv_open(samba, local);
    pthread_mutex_unlock(&m_config);
    return cd;
}

iconv_t samba2local(){
    char	*local = local_charset;
    char	*samba = samba_charset;
    iconv_t	cd;
    
    pthread_mutex_lock(&m_config);
    if (opt.local_charset != NULL) local = opt.local_charset;
    if (opt.samba_charset != NULL) samba = opt.samba_charset;
    cd = iconv_open(local, samba);
    pthread_mutex_unlock(&m_config);
    return cd;
}

void fill_login_info(char *user, int userlen, char *passwd, int passwdlen){
    char	*username = login;
    char	*password = "";
    
    pthread_mutex_lock(&m_config);
    if (opt.username != NULL) username = opt.username;
    if (opt.password != NULL) password = opt.password;
    strncpy(user, username, userlen - 1); user[userlen - 1] = 0;
    strncpy(passwd, password, passwdlen - 1); passwd[passwdlen - 1] = 0;
    pthread_mutex_unlock(&m_config);
}

void debug_print(const char *fmt, ...){
    va_list ap;
    
    pthread_mutex_lock(&m_config);
    if (stdlog == NULL){
	if (opt.log_file != NULL) stdlog = fopen(opt.log_file, "a");
	if (stdlog == NULL) stdlog = stderr;
    }
    va_start(ap, fmt);
    vfprintf(stdlog, fmt, ap);
    fflush(stdlog);
    va_end(ap);
    pthread_mutex_unlock(&m_config);
}
