#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <errno.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/vfs.h>
#include <time.h>
#include <iconv.h>
#include <libsmbclient.h>
#include "config.h"
#include "convert.h"
#include "smbdescr.h"
#include "smbitem.h"
#include "update.h"
#include "function.h"

static int samba_getattr(const char *path, struct stat *stbuf){
    int				error = 0;
    const char			*smbpath, *name;
    int				pathtype, name_len, lnklen, block_size;
    struct SmbDescriptor	*descr;
    void			*data;
    
    DPRINT("(%s)\n", path);
    
    pathtype = GetPathType(path, &data);
    block_size = option(OPT_BLOCK_SIZE);
    switch(pathtype){
	case PATH_UNKNOWN:
	    return -EINVAL;
	case PATH_TO_LINK:
	    if (data != NULL){
		lnklen = strlen((char*)data);
		free(data);
	    }else lnklen = 0;
	    memset(stbuf, 0, sizeof(struct stat));
	    stbuf->st_mode = 0120777;	    /* protection */
	    stbuf->st_nlink = 1;	    /* number of hard links */
	    stbuf->st_uid = 0;		    /* user ID of owner */
	    stbuf->st_gid = 0;		    /* group ID of owner */
	    stbuf->st_size = lnklen;	    /* total size, in bytes */
	    stbuf->st_blksize = block_size; /* blocksize for filesystem I/O */
	    return 0;
	case PATH_TO_DIR:
	    if (data != NULL) ReleaseSmbItemArray((struct smbitemarray*)data);
	case PATH_TO_COMP:
	case PATH_TO_SHARE:
	    memset(stbuf, 0, sizeof(struct stat));
	    stbuf->st_mode = 0040777;	    /* protection */
	    stbuf->st_nlink = 2;	    /* number of hard links */
	    stbuf->st_uid = 0;		    /* user ID of owner */
	    stbuf->st_gid = 0;		    /* group ID of owner */
	    stbuf->st_size = 0;		    /* total size, in bytes */
	    stbuf->st_blksize = block_size; /* blocksize for filesystem I/O */
	    return 0;
	default:
	    if ((smbpath = path2samba(path)) == NULL) return -ENOMEM;
	    name = GetFirstPath(path, &name_len);
	    if ((descr = GetSmbDescriptor(name, name_len)) == NULL){
		error = -ENOMEM;
		goto error0;
	    }
	    if (descr->ctx->stat(descr->ctx, smbpath, stbuf) != 0)
		error = -(descr->error = errno);
	    else stbuf->st_blksize = block_size; /* blocksize for filesystem I/O */
 	    ReleaseSmbDescriptor(descr);
	  error0:
	    free((void*)smbpath);
	    return error;
    }
}

static int samba_readlink(const char *path, char *buf, size_t size){
    void	*data;

    DPRINT("(%s, %d)\n", path, size);

    switch(GetPathType(path, &data)){
	case PATH_TO_LINK:
	    if (data == NULL) return -EINVAL;
	    strncpy(buf, (char*)data, size);
	    free(data);
	    return 0;
	case PATH_TO_DIR:
	    if (data != NULL) ReleaseSmbItemArray((struct smbitemarray*)data);
	default: 
	    return -EINVAL;
    }
}

static int samba_mknod(const char *path, mode_t mode, dev_t rdev){
    int				name_len, error = 0;
    const char			*smbpath, *name;
    struct SmbDescriptor	*descr;
    
    (void) rdev;

    DPRINT("(%s, %o)\n", path, mode);

    if (GetPathType(path, NULL) != PATH_TO_SMBFILE) return -EPERM;
    if ((mode & S_IFMT) != S_IFREG) return -EPERM;
    if ((smbpath = path2samba(path)) == NULL) return -ENOMEM;
    name = GetFirstPath(path, &name_len);
    if ((descr = GetSmbDescriptor(name, name_len)) == NULL){
	free((void*)smbpath);
	return -ENOMEM;
    }
    if ((descr->fd = descr->ctx->creat(descr->ctx, smbpath, mode)) == NULL)
	error = -(descr->error = errno);
    else descr->ctx->close(descr->ctx, descr->fd);
    ReleaseSmbDescriptor(descr);
    free((void*)smbpath);
    return error;
}

static int samba_mkdir(const char *path, mode_t mode){
    int				name_len, error = 0;
    const char			*smbpath, *name;
    struct SmbDescriptor	*descr;

    DPRINT("(%s, %o)\n", path, mode);
    
    if (GetPathType(path, NULL) != PATH_TO_SMBFILE) return -EPERM;
    if ((smbpath = path2samba(path)) == NULL) return -ENOMEM;
    name = GetFirstPath(path, &name_len);
    if ((descr = GetSmbDescriptor(name, name_len)) == NULL){
	free((void*)smbpath);
	return -ENOMEM;
    }
    if (descr->ctx->mkdir(descr->ctx, smbpath, mode) != 0)
	error = -(descr->error = errno);
    ReleaseSmbDescriptor(descr);
    free((void*)smbpath);
    return error;
}

static int samba_unlink(const char *path){
    int				name_len, error = 0;
    const char			*smbpath, *name;
    struct SmbDescriptor	*descr;
    
    DPRINT("(%s)\n", path);
    
    if (GetPathType(path, NULL) != PATH_TO_SMBFILE) return -EPERM;
    if ((smbpath = path2samba(path)) == NULL) return -ENOMEM;
    name = GetFirstPath(path, &name_len);
    if ((descr = GetSmbDescriptor(name, name_len)) == NULL){
	free((void*)smbpath);
	return -ENOMEM;
    }
    if (descr->ctx->unlink(descr->ctx, smbpath) != 0)
	error = -(descr->error = errno);
    ReleaseSmbDescriptor(descr);
    free((void*)smbpath);
    return error;
}

static int samba_rmdir(const char *path){
    int				name_len, error = 0;
    const char			*smbpath, *name;
    struct SmbDescriptor	*descr;
    
    DPRINT("(%s)\n", path);
    
    if (GetPathType(path, NULL) != PATH_TO_SMBFILE) return -EPERM;
    if ((smbpath = path2samba(path)) == NULL) return -ENOMEM;
    name = GetFirstPath(path, &name_len);
    if ((descr = GetSmbDescriptor(name, name_len)) == NULL){
	free((void*)smbpath);
	return -ENOMEM;
    }
    if (descr->ctx->rmdir(descr->ctx, smbpath) != 0)
	error = -(descr->error = errno);
    ReleaseSmbDescriptor(descr);
    free((void*)smbpath);
    return error;
}

static int samba_symlink(const char *from, const char *to){
    (void) from;
    (void) to;
    
    DPRINT("(%s, %s)\n", from, to);
    
    return -EPERM;
}

static int samba_rename(const char *from, const char *to){
    int				name_len, error = 0;
    const char			*smbfrom, *smbto, *name;
    struct SmbDescriptor	*descr;
    
    DPRINT("(%s, %s)\n", from, to);
    
    if ((GetPathType(from, NULL) != PATH_TO_SMBFILE) ||
	(GetPathType(to, NULL) != PATH_TO_SMBFILE)) return -EPERM;
    if ((smbfrom = path2samba(from)) == NULL){
	error = -ENOMEM;
	goto error0;
    }
    if ((smbto = path2samba(to)) == NULL){
	error = -ENOMEM;
	goto error1;
    }
    name = GetFirstPath(from, &name_len);
    if ((descr = GetSmbDescriptor(name, name_len)) == NULL){
	error = -ENOMEM;
	goto error2;
    }
    if (descr->ctx->rename(descr->ctx, smbfrom, descr->ctx, smbto) != 0)
	error = -(descr->error = errno);
    ReleaseSmbDescriptor(descr);
  error2:
    free((void*)smbto);
  error1:    
    free((void*)smbfrom);
  error0:
    return error;
}

static int samba_link(const char *from, const char *to){
    (void) from;
    (void) to;
    
    DPRINT("(%s, %s)\n", from, to);
    
    return -EPERM;
}

static int samba_chmod(const char *path, mode_t mode){
    int				name_len, error = 0;
    const char			*smbpath, *name;
    struct SmbDescriptor	*descr;
    
    DPRINT("(%s, %o)\n", path, mode);
    
    if (GetPathType(path, NULL) != PATH_TO_SMBFILE) return -EPERM;
    if ((smbpath = path2samba(path)) == NULL) return -ENOMEM;
    name = GetFirstPath(path, &name_len);
    if ((descr = GetSmbDescriptor(name, name_len)) == NULL){
	free((void*)smbpath);
	return -ENOMEM;
    }
    if (descr->ctx->chmod(descr->ctx, smbpath, mode) != 0)
	error = -(descr->error = errno);
    ReleaseSmbDescriptor(descr);
    free((void*)smbpath);
    return error;
}

static int samba_chown(const char *path, uid_t uid, gid_t gid){
    (void) path;
    (void) uid;
    (void) gid;

    DPRINT("(%s, uid=%d, gid=%d)\n", path, uid, gid);
    if (option(OPT_QUIET_FLAG)) return 0;
    else return -EPERM;
}

static int samba_truncate(const char *path, off_t size){
    int				name_len, error = 0;
    const char			*smbpath, *name;
    struct SmbDescriptor	*descr;

    DPRINT("(%s, %d)\n", path, (int) size);
    
    if (size != 0) return -EACCES;
    if (GetPathType(path, NULL) != PATH_TO_SMBFILE) return -EPERM;
    if ((smbpath = path2samba(path)) == NULL) return -ENOMEM;
    name = GetFirstPath(path, &name_len);
    if ((descr = GetSmbDescriptor(name, name_len)) == NULL){
	free((void*)smbpath);
	return -ENOMEM;
    }
    if ((descr->fd = descr->ctx->creat(descr->ctx, smbpath, 0777)) == NULL)
	error = -(descr->error = errno);
    else descr->ctx->close(descr->ctx, descr->fd);
    ReleaseSmbDescriptor(descr);
    free((void*)smbpath);
    return error;
}

static int samba_utime(const char *path, struct utimbuf *buf){
    int				name_len, error = 0;
    const char			*smbpath, *name;
    struct SmbDescriptor	*descr;
    struct timeval		tbuf;
    
    DPRINT("(%s, %u)\n", path, (unsigned int)buf->modtime);
    
    if (GetPathType(path, NULL) != PATH_TO_SMBFILE) return -EACCES;
    if ((smbpath = path2samba(path)) == NULL) return -ENOMEM;
    name = GetFirstPath(path, &name_len);
    if ((descr = GetSmbDescriptor(name, name_len)) == NULL){
	free((void*)smbpath);
	return -ENOMEM;
    }
    tbuf.tv_sec = buf->modtime;
    tbuf.tv_usec = 0;
    if (descr->ctx->utimes(descr->ctx, smbpath, &tbuf) != 0)
	error = -(descr->error = errno);
    ReleaseSmbDescriptor(descr);
    free((void*)smbpath);
    return error;
}

static int samba_open(const char *path, struct fuse_file_info *fi){
    int				name_len, error = 0;
    const char			*smbpath, *name;
    struct SmbDescriptor	*descr;

    DPRINT("(%s, flags=%o, fh=%d)\n", path, fi->flags, (int)fi->fh);
    if (GetPathType(path, NULL) != PATH_TO_SMBFILE) return -EACCES;
    if ((smbpath = path2samba(path)) == NULL) return -ENOMEM;
    name = GetFirstPath(path, &name_len);
    descr = GetSmbDescriptor(name, name_len);
    if (descr != NULL)
	if ((descr->fd = descr->ctx->open(descr->ctx, smbpath, 
			fi->flags, 0777)) == NULL){
	    error = -(descr->error = errno);
	    ReleaseSmbDescriptor(descr);
	    descr = NULL;
	}
    fi->fh = (unsigned long) descr;	
    free((void*)smbpath);
    return error;
}

static int samba_flush(const char *path, struct fuse_file_info *fi){

    DPRINT("(%s, fh=%d, flags=%o)\n", path, (int)fi->fh, fi->flags);
        
    return 0;
}

static int samba_release(const char *path, struct fuse_file_info *fi){
    int				error = 0;
    struct SmbDescriptor	*descr;

    DPRINT("(%s, fh=%d, flags=%o)\n", path, (int)fi->fh, fi->flags);

    if ((descr = (struct SmbDescriptor*) fi->fh) == NULL) return -EBADF;
    if (descr->fd == NULL) error = -EBADF;
    else if (descr->ctx->close(descr->ctx, descr->fd) != 0)
	error = -(descr->error = errno); 
    ReleaseSmbDescriptor(descr);
    return error;
}

static int samba_reopen(const char *path, struct fuse_file_info *fi){
    int				error = 0;
    const char			*smbpath;
    struct SmbDescriptor	*descr;

    DPRINT("(%s, flags=%o, fh=%d)\n", path, fi->flags, (int)fi->fh);
    if ((descr = (struct SmbDescriptor*) fi->fh) == NULL) return -EBADF;
    if (descr->fd == NULL) error = -EBADF;
    descr->ctx->close(descr->ctx, descr->fd);
    descr->fd = NULL;
    descr->offset = 0;
    descr->error = 0;
    
    if ((smbpath = path2samba(path)) == NULL) return -ENOMEM;
    if ((descr->fd = descr->ctx->open(descr->ctx, smbpath, 
	fi->flags, 0777)) == NULL) error = -(descr->error = errno);
    free((void*)smbpath);
    return error;
}

static int samba_read(const char *path, char *buf, size_t size, off_t offset,
		    struct fuse_file_info *fi){
    int				block_size, result, err_count = 0;
    struct SmbDescriptor	*descr;

    DPRINT("(%s, %d, fh=%d, offset=%u, flags=%o)\n", path, size, (int)fi->fh, (unsigned int)offset, fi->flags);

  restart:
    result = 0;
    block_size = option(OPT_BLOCK_SIZE);
    if ((descr = (struct SmbDescriptor*) fi->fh) == NULL) return -EBADF;
    if (descr->fd == NULL) return -EBADF;
    if (descr->offset != offset)
	if ((descr->offset = descr->ctx->lseek(descr->ctx, descr->fd,
	    offset, SEEK_SET)) == (off_t) -1) goto retry;

    while(size > 0){
	int	res, bytes = size;
	if (bytes > block_size) bytes = block_size;
	if ((res = descr->ctx->read(descr->ctx, descr->fd,
	    buf + result, bytes)) == (ssize_t) -1) goto retry;
	size -= res; result += res;
	descr->offset += res;
	if (res != bytes) break;
    }

  error0:
    return result;

  retry:
    result = -(descr->error = errno);
    switch(errno){
	case EINVAL:
	case EBADF:
	    if (err_count > 0) goto error0;
	    err_count++;
	    if (samba_reopen(path, fi) == 0) goto restart;
//	    samba_release(path, fi);
//	    if (samba_open(path, fi) == 0) goto restart;
	default:
	    goto error0;
    }
}

static int samba_write(const char *path, const char *buf, size_t size,
		    off_t offset, struct fuse_file_info *fi){
    int				block_size, result, err_count = 0;
    struct SmbDescriptor	*descr;

    DPRINT("(%s, %d, fh=%d, offset=%u, flags=%o)\n", path, size, (int)fi->fh, (unsigned int)offset, fi->flags);

  restart:
    result = 0;
    block_size = option(OPT_BLOCK_SIZE);
    if ((descr = (struct SmbDescriptor*) fi->fh) == NULL) return -EBADF;
    if (descr->fd == NULL) return -EBADF;
    if (descr->offset != offset)
	if ((descr->offset = descr->ctx->lseek(descr->ctx, descr->fd,
	    offset, SEEK_SET)) == (off_t) -1) goto retry;

    while(size > 0){
	int	res, bytes = size;
	if (bytes > block_size) bytes = block_size;
	if ((res = descr->ctx->write(descr->ctx, descr->fd,
	    (char*)buf + result, bytes)) == (ssize_t) -1) goto retry;
	size -= res; result += res;
	descr->offset += res;
	if (res != bytes) break;
    }
  error0:
    return result;

  retry:
    result = -(descr->error = errno);
    switch(errno){
	case EINVAL:
	case EBADF:
	    if (err_count > 0) goto error0;
	    err_count++;
	    if (samba_reopen(path, fi) == 0) goto restart;
//	    samba_release(path, fi);
//	    if (samba_open(path, fi) == 0) goto restart;
	default:
	    goto error0;
    }
}

static int samba_statfs(const char *path, struct statfs *stbuf){
    DPRINT("(%s)\n", path);

    memset(stbuf, 0, sizeof(struct statfs));
    stbuf->f_namelen = FILENAME_MAX;
    return 0;
}


static int samba_fsync(const char *path, int isdatasync,
			struct fuse_file_info *fi){
    (void)isdatasync;
    
    DPRINT("(%s, fh=%d, flags=%o)\n", path, (int)fi->fh, fi->flags);
    
    return 0;
}

static int samba_opendir(const char *path, struct fuse_file_info *fi){
    int				name_len, error = 0;
    const char			*smbpath, *name;
    struct SmbDescriptor	*descr;

    DPRINT("(%s, flags=%o, fh=%d)\n", path, fi->flags, (int)fi->fh);
    switch(GetPathType(path, NULL)){
	case PATH_TO_DIR:
	    fi->fh = (unsigned int) NULL;
	    return 0;
	case PATH_TO_COMP:
	case PATH_TO_SHARE:
	case PATH_TO_SMBFILE:
	    break;
	case PATH_TO_LINK:
	case PATH_UNKNOWN:
	default:
	    return -ENOTDIR;
    }
    if ((smbpath = path2samba(path)) == NULL) return -ENOMEM;
    name = GetFirstPath(path, &name_len);
    if ((descr = GetSmbDescriptor(name, name_len)) == NULL) error = -ENOMEM;
    else if ((descr->fd = descr->ctx->opendir(descr->ctx, smbpath)) == NULL){
	error = -(descr->error = errno);
	ReleaseSmbDescriptor(descr);
	descr = NULL;
    }
    fi->fh = (unsigned long) descr;
    free((void*)smbpath);
    return error;
}

static int samba_releasedir(const char *path, struct fuse_file_info *fi){
    int				error = 0;
    struct SmbDescriptor	*descr;

    DPRINT("(%s, fh=%d, flags=%o)\n", path, (int)fi->fh, fi->flags);
    if ((descr = (struct SmbDescriptor*) fi->fh) != NULL){
	if (descr->fd == NULL) error = -EBADF;
	else if (descr->ctx->close(descr->ctx, descr->fd) != 0)
	    error = -(descr->error = errno);
	ReleaseSmbDescriptor(descr);
    }
    return error;
}

static int samba_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
			off_t offset, struct fuse_file_info *fi){
    int				i, grplen;
    const char			*name, *group;
    char			*filename, *link;
    struct smbc_dirent		*dirent;
    struct SmbDescriptor	*descr;
    struct smbitemarray		*array;
    struct stat 		st;
    void			*data;
    
    DPRINT("(%s)\n", path);

    (void) offset;    

    memset(&st, 0, sizeof(st));
    descr = (struct SmbDescriptor*) fi->fh;
    if (descr == NULL)
	switch(GetPathType(path, &data)){
	    case PATH_TO_DIR:
		array = (struct smbitemarray*) data;

		st.st_mode = S_IFDIR;
		if (filler(buf, ".",  &st, 0)) goto error0;
		if (filler(buf, "..", &st, 0)) goto error0;
		if (array == NULL) goto error0;

		for(i = 0; i < array->count; i++){
		    if (array->elem[i]->type == SMBITEM_LINK)
			st.st_mode = S_IFLNK;
		    else st.st_mode = S_IFDIR;
		    if (filler(buf, array->elem[i]->name, &st, 0)) break;
		}
	      error0:
		if (array != NULL) ReleaseSmbItemArray(array);
		return 0;
	    case PATH_TO_LINK:
		if (data != NULL) free(data);
	    default:
		return -EBADF;
	}

    if (descr->fd == NULL) return -EBADF;
    st.st_mode = S_IFDIR;
    if (filler(buf, ".",  &st, 0)) return 0;
    if (filler(buf, "..", &st, 0)) return 0;

    // this is safe, couse cd will be checked in convert function.
    iconv_t cd = samba2local();
    while((dirent = descr->ctx->readdir(descr->ctx, descr->fd)) != NULL){
	if (strcmp(dirent->name, ".") == 0) continue;
	if (strcmp(dirent->name, "..") == 0) continue;
	if ((name = from_samba(dirent->name, cd)) == NULL) continue;
	switch(dirent->smbc_type){
	    case SMBC_WORKGROUP:
		goto error1;
	    case SMBC_SERVER:
		group = GetFirstPath(path, &grplen);

	    	if ((filename = malloc(grplen + strlen(name) + 2)) == NULL)
		    goto error2;
		strncpy(filename, group, grplen);
		filename[grplen] = '\0';
		mkgroup(filename, AUTOPATH);
		
		filename[grplen] = '/';
		strcpy(filename + grplen + 1, name);

		if ((link = malloc(strlen(name) + 4)) == NULL) goto error3;
		strcpy(link, "../");
		strcpy(link + 3, name);

		mklink(filename, link, AUTOPATH);

		free(link);
	      error3:
	        free(filename);
	      error2:

		st.st_mode = S_IFDIR;
		break;
	    case SMBC_FILE_SHARE:
		st.st_mode = S_IFDIR;
		break;
	    case SMBC_PRINTER_SHARE:
	    case SMBC_COMMS_SHARE:
	    case SMBC_IPC_SHARE:
		goto end;
	    case SMBC_DIR:
		st.st_mode = S_IFDIR;
		break;
	    case SMBC_FILE:
		st.st_mode = S_IFREG;
		break;
	    case SMBC_LINK:
		goto end;
	    default:
		goto end;
	}
	if (filler(buf, name, &st, 0)) break;
      end:
	free((char*)name);
    }
  error1:
    iconv_close(cd);
    return -descr->error;
}

static void* samba_init(){
    DPRINT("\n");
    
    if (InitThreads() != 0){
	printf("Could not create new thread\n");
	exit(1);
    }	
    return NULL;
}

static void samba_destroy(void *private_data){
    (void)private_data;

    DPRINT("\n");
    DestroyThreads();
}
		
struct fuse_operations smb_oper = {
    .getattr	= samba_getattr,
    .readlink	= samba_readlink,
    .mknod	= samba_mknod,
    .mkdir	= samba_mkdir,
    .unlink	= samba_unlink,
    .rmdir	= samba_rmdir,
    .symlink	= samba_symlink,
    .rename	= samba_rename,
    .link	= samba_link,
    .chmod	= samba_chmod,
    .chown	= samba_chown,
    .truncate	= samba_truncate,
    .utime	= samba_utime,
    .open	= samba_open,
    .read	= samba_read,
    .write	= samba_write,
    .statfs	= samba_statfs,
    .flush	= samba_flush,
    .release	= samba_release,
    .fsync	= samba_fsync,
    .opendir	= samba_opendir,
    .readdir	= samba_readdir,
    .releasedir	= samba_releasedir,
    .init	= samba_init,
    .destroy	= samba_destroy,
//    .setxattr	= samba_setxattr,
//    .getxattr	= samba_getxattr,
//    .listxattr	= samba_listxattr,
//    .removexattr= samba_removexattr,    
};

