#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/types.h>
#include <libsmbclient.h>
#include "config.h"
#include "smbdescr.h"

struct SmbDescrList{
    struct SmbDescriptor	*first;
    struct SmbDescriptor	*last;
};

int			descr_count	= 0;
struct SmbDescrList	free_list	= { NULL, NULL };
struct SmbDescrList	busy_list	= { NULL, NULL };
pthread_mutex_t		m_smbdescr	= PTHREAD_MUTEX_INITIALIZER;

void smbc_auth_fn(
		const char	*server,
		const char	*share,
		char		*wrkgrp, int wrkgrplen,
		char		*user,   int userlen,
		char		*passwd, int passwdlen){

    (void) server;
    (void) share;
    (void) wrkgrp;
    (void) wrkgrplen;

    fill_login_info(user, userlen, passwd, passwdlen);
}

void print_descr(char *head){
    int				free_cnt = 0, busy_cnt = 0;
    struct SmbDescriptor 	*descr;
    
    fprintf(stderr, "\n%s\nfree list: ", head);
    descr = free_list.first;
    while(descr != NULL){
	free_cnt++;
	fprintf(stderr, "<%s> ", descr->name);
	descr = descr->next;
    };
    fprintf(stderr, "\nbusy list : ");
    descr = busy_list.first;
    while(descr != NULL){
	busy_cnt++;
	fprintf(stderr, "<%s> ", descr->name);
	descr = descr->next;
    };
    fprintf(stderr, "\nTotal: free=%d, busy=%d, total=%d\n---------------\n ",
	free_cnt, busy_cnt, free_cnt + busy_cnt);
}    

struct SmbDescriptor* NewSmbDescriptor(const char *name, int name_len){
    struct SmbDescriptor	*descr;

    if ((descr = (struct SmbDescriptor*)malloc(sizeof(struct SmbDescriptor) +
	    name_len + 1)) == NULL) goto error0;
    memset(descr, 0, sizeof(struct SmbDescriptor) + name_len + 1);
    strncpy(descr->name, name, name_len);
    
    if ((descr->ctx = smbc_new_context()) == NULL) goto error1;
    descr->ctx->debug = option(OPT_DEBUG);
    descr->ctx->callbacks.auth_fn = smbc_auth_fn;
    if (smbc_init_context(descr->ctx) == NULL) goto error2;
    
    descr_count++;
    DPRINT("name=%s, descr_count=%d\n", descr->name, descr_count);
    return descr;
  error2:
    smbc_free_context(descr->ctx, 1);
  error1:
    free(descr);
  error0:
    return NULL;
}

void CleanupSmbDescriptor(struct SmbDescriptor *descr){
    if (descr->fd != NULL) descr->ctx->close(descr->ctx, descr->fd);
    descr->fd = NULL; descr->offset = 0; descr->error = 0;
}

void DeleteSmbDescriptor(struct SmbDescriptor *descr){
    DPRINT("name=%s, descr_count=%d\n", descr->name, descr_count);
    CleanupSmbDescriptor(descr);
    smbc_free_context(descr->ctx, 1);
    free(descr);
    descr_count--;
}

int IsGoodSmbDescriptor(struct SmbDescriptor *descr){
    switch (descr->error){
	case 0:
	case EACCES:
	case EBADF:
	case EBUSY:
	case EEXIST:
	case EISDIR:
	case ENODEV:
	case ENOENT:
	case ENOTDIR:
	case ENOTEMPTY:
	case ENOTSUP:
	case EPERM:
	case EXDEV:	return 1;
	
	case EINVAL:
	case ENOMEM:
	default:	return 0;
    }
}

void add_to_descr_list(struct SmbDescrList *list,
			struct SmbDescriptor *elem){
    elem->next = list->first;
    elem->prev = NULL;
    if (elem->next != NULL) elem->next->prev = elem;
    list->first = elem;
    if (list->last == NULL) list->last = elem;
}

void remove_from_descr_list(struct SmbDescrList *list,
			struct SmbDescriptor *elem){
    if (elem->prev != NULL) elem->prev->next = elem->next; 
    if (elem->next != NULL) elem->next->prev = elem->prev;
    if (elem == list->first) list->first = elem->next;
    if (elem == list->last)  list->last  = elem->prev;
}

struct SmbDescriptor* find_in_descr_list(struct SmbDescrList *list,
					const char *name, int name_len){
    struct SmbDescriptor  *elem = list->first;
    while(elem != NULL){
	if ((strlen(elem->name) == (size_t) name_len) &&
	    (strncasecmp(elem->name, name, name_len) == 0)) return elem;
	elem = elem->next;
    }
    return NULL;
}

void delete_descr_list(struct SmbDescrList *list){
    struct SmbDescriptor   *tmp, *elem = list->first;
    while(elem != NULL){
	tmp = elem; elem = elem->next;
	DeleteSmbDescriptor(tmp);
    }
    list->first = NULL;
    list->last = NULL;
}

struct SmbDescriptor* GetSmbDescriptor(const char *name, int name_len){
    int			 descr_max;
    struct SmbDescriptor *descr;

    DPRINT("name[%d]=%s\n", name_len, name);

    descr_max = option(OPT_SMB_MAX_DESCR);
    pthread_mutex_lock(&m_smbdescr);
    print_descr("--- get descr 1 ----");
    descr = find_in_descr_list(&free_list, name, name_len);
    if (descr != NULL) remove_from_descr_list(&free_list, descr);
    else{
	while((descr_count >= descr_max) && (free_list.last != NULL)){
	    descr = free_list.last;
	    remove_from_descr_list(&free_list, descr);
	    DeleteSmbDescriptor(descr);
	}
	if (descr_count >= descr_max) descr = NULL;
	else descr = NewSmbDescriptor(name, name_len);
    }
    if (descr != NULL){
	add_to_descr_list(&busy_list, descr);
	CleanupSmbDescriptor(descr);
    }
    print_descr("--- get descr 2 ----");
    pthread_mutex_unlock(&m_smbdescr);
    return descr;
}

void ReleaseSmbDescriptor(struct SmbDescriptor *descr){
    int descr_max = option(OPT_SMB_MAX_DESCR);
    
    CleanupSmbDescriptor(descr);
    pthread_mutex_lock(&m_smbdescr);
    print_descr("--- release descr 1 ----");
    fprintf(stderr, "error = %d\n", descr->error);
    remove_from_descr_list(&busy_list, descr);
    if ((IsGoodSmbDescriptor(descr)) && (descr_count <= descr_max))
	add_to_descr_list(&free_list, descr);
    else DeleteSmbDescriptor(descr);
    print_descr("--- release descr 2 ----");
    pthread_mutex_unlock(&m_smbdescr);
    return;
}

void DestroyContextList(){
    pthread_mutex_lock(&m_smbdescr);
    delete_descr_list(&free_list);
    delete_descr_list(&busy_list);
    pthread_mutex_unlock(&m_smbdescr);
}
