#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <pthread.h>
#include "config.h"
#include "smbitem.h"

pthread_mutex_t 	msmbitem	= PTHREAD_MUTEX_INITIALIZER;
struct smbitemarray	user_tree	= {0, 0, NULL};
struct smbitemarray	auto_tree	= {0, 0, NULL};

void DestroySmbItemArray(struct smbitemarray *array);

struct smbitem* MakeItem(const char *name, int type, const char *link){
    int			alloc_len;
    struct smbitem	*item;
    
    if (name == NULL) return NULL;

    switch(type){
	case SMBITEM_COMP:
	    alloc_len = 0;
	    break;
	case SMBITEM_LINK:
	    if (link == NULL) return NULL;
	    alloc_len = strlen(link) + 1;
	    break;
	case SMBITEM_GROUP:
	    alloc_len = sizeof(struct smbitemarray);
	    break;
	default:
	    return NULL;
    };
    alloc_len += sizeof(struct smbitem) + strlen(name) + 1;
    
    if ((item = malloc(alloc_len)) == NULL) return NULL;
    memset(item, 0, alloc_len);
    item->type = type;
    item->ctime = time(NULL);

    item->name = item->data;
    if (type == SMBITEM_LINK){
	strcpy(item->data, link);
	item->name += strlen(link) + 1;
    }else if (type == SMBITEM_GROUP)
	item->name += sizeof(struct smbitemarray);
    strcpy((char*)item->name, name);
    return item;
}

void DestroySmbItem(struct smbitem *item){
    if (item->type == SMBITEM_GROUP)
	DestroySmbItemArray((struct smbitemarray*) item->data);
    free(item);
}    

void RequestSmbItem(struct smbitem *item){
    if (item == NULL) return;
    if (item->ref_count < 0){
	DPRINT("Error: item->ref_count < 0\n");
	return;
    }
    item->ref_count++;
}

void ReleaseSmbItem(struct smbitem *item){
    if (item == NULL) return;
    item->ref_count--;
    if (item->ref_count < 0){
	DPRINT("Error: item->ref_count < 0\n");
	return;
    }
    if (item->ref_count == 0) DestroySmbItem(item);
}


void DestroySmbItemArray(struct smbitemarray *array){
    int			i;

    if (array == NULL) return;
    if (array->count < 0){
	DPRINT("Error: array->count < 0\n");
	return;
    }
    if (array->max_count < array->count){
	DPRINT("Error: array->max_count < array->count\n");
	return;
    }
    if ((array->elem == NULL) && (array->max_count != 0)){
	DPRINT("Error: array->elem = NULL, but array->max_count != 0\n");
	return;
    }
    for(i = 0; i < array->count; i++) ReleaseSmbItem(array->elem[i]);
    if (array->elem != NULL) free(array->elem);
    memset(array, 0, sizeof(struct smbitemarray));
}

int SmbItemArrayAdd(struct smbitemarray *array, struct smbitem *item){
    struct smbitem      **newelem;
    int			max_count;
    
    max_count = array->max_count;
    if (array->count == max_count){
	max_count += 32;
	if ((newelem = malloc(max_count * sizeof(struct smbitem*))) == NULL)
	    return 0;
	memset(newelem, 0, max_count * sizeof(struct smbitem*));
	if (array->elem != NULL){
	    memcpy(newelem, array->elem, 
		array->count * sizeof(struct smbitem*));
	    free(array->elem);
	}
	array->elem = newelem;
	array->max_count = max_count;
    }
    RequestSmbItem(item);
    array->elem[array->count] = item;	
    array->count++;
    return 1;
}

int SmbItemArrayReplace(struct smbitemarray *array, int i, struct smbitem *item){
    if ((i < 0) || (i >= array->count)) return 0;
    ReleaseSmbItem(array->elem[i]);
    RequestSmbItem(item);
    array->elem[i] = item;
    return 1;
}

int SmbItemArrayDel(struct smbitemarray *array, int i){
    if ((i < 0) || (i >= array->count)) return 0;
    ReleaseSmbItem(array->elem[i]);
    array->count--;
    if (i != array->count) memmove(&array->elem[i], &array->elem[i + 1],
	(array->count - i) * sizeof(struct smbitem*));
    return 1;
}

int FindSmbItem(struct smbitemarray *array, const char *name){
    int i;
    
    for(i = 0; i < array->count; i++)
	if (strcasecmp(array->elem[i]->name, name) == 0) return i;
    return -1;
}

int FindSmbItemLen(struct smbitemarray *array, const char *name, int len){
    int i;
    
    for(i = 0; i < array->count; i++)
	if ((strlen(array->elem[i]->name) == (size_t) len) &&
	    (strncasecmp(array->elem[i]->name, name, len) == 0)) return i;
    return -1;
}

int mkitem(struct smbitemarray *array, char *name, int type, char *link){
    int			i;
    struct smbitem      *item;
    
    if ((i = FindSmbItem(array, name)) >= 0){
	if (array->elem[i]->type == type){
	    if ((type != SMBITEM_LINK) ||
		(strcmp(array->elem[i]->data, link) == 0)){
		
		array->elem[i]->ctime = time(NULL);
		return 1;
	    }
	}
    }

    if ((item = MakeItem(name, type, link)) == NULL) return 0;
    if (i < 0){
	if (SmbItemArrayAdd(array, item)) return 1;
    }else{
	if (SmbItemArrayReplace(array, i, item)) return 1;
    }
    DestroySmbItem(item);
    return 0;
}

struct smbitemarray* finddir(struct smbitemarray *array, char *path){
    int i;

    if (strcmp(path, "") != 0){
	if ((i = FindSmbItem(array, path)) < 0) return NULL;
	if (array->elem[i]->type != SMBITEM_GROUP) return NULL;
	array = (struct smbitemarray*)array->elem[i]->data;
    }
    return array;
}

struct smbitemarray* finddirlen(struct smbitemarray *array, char *path, int len){
    int i;

    if (len != 0){
	if ((i = FindSmbItemLen(array, path, len)) < 0) return NULL;
	if (array->elem[i]->type != SMBITEM_GROUP) return NULL;
	array = (struct smbitemarray*)array->elem[i]->data;
    }
    return array;
}

void getarray(struct smbitemarray *array, struct smbitemarray *result){
    int			i, j;
    
    for(i = 0; i < array->count; i++){
	if ((j = FindSmbItem(result, array->elem[i]->name)) >= 0)
	    SmbItemArrayReplace(result, j, array->elem[i]);
	else SmbItemArrayAdd(result, array->elem[i]);
    }
}

void killold(struct smbitemarray *array, time_t ctime){
    struct smbitemarray	*group;
    int			i;
    
    i = 0;
    while(i < array->count){
	if (array->elem[i]->type == SMBITEM_GROUP){
	    group = (struct smbitemarray*)array->elem[i]->data;
	    killold(group, ctime);
	    if (group->count > 0){
		i++;
		continue;
	    }
	}
	if (array->elem[i]->ctime >= ctime){
	    i++;
	    continue;
	}
	SmbItemArrayDel(array, i);
    }
}

int FindSlash(const char *path){
    int i, len; 

    if (path == NULL) return 0;
    len = strlen(path);
    for(i = 0; i < len; i++) if (path[i] == '/') return 1;
    return 0;
}

int GetPathComponent(const char *path){
    int mode = 0, count = 0;

    if (path == NULL) return 0;
    // find a number of component in a path
    for( ; *path; path++)
	if (*path == '/') mode = 0;
	else if (mode == 0){
	    mode = 1; if (++count >= 3) return 3;
	}
    return count;
}

const char* GetFirstPath(const char* path, int *result_len){
    for( ; *path; path++) if (*path != '/') break;
    for(*result_len = 0; (path[*result_len] != '\0') &&
			 (path[*result_len] != '/'); (*result_len)++);
    return path;
}

const char* GetSecondPath(const char* path, int *result_len){
    for( ; *path; path++) if (*path != '/') break;
    for( ; *path; path++) if (*path == '/') break;
    return GetFirstPath(path, result_len);
}

int mkgroup(char *path, int flag){
    struct smbitemarray *array = NULL;
    int			result;
    
    DPRINT("path=%s, flag=%d\n", path, flag);
    path = (char*) GetFirstPath(path, &result);
    if (FindSlash(path)) return 0;
    if (strcmp(path, "") == 0) return 0;
    if (strcmp(path, ".") == 0) return 0;
    if (strcmp(path, "..") == 0) return 0;
    if (flag & AUTOPATH) array = &auto_tree;
    if (flag & USERPATH) array = &user_tree;
    if (array == NULL) return 0;    

    pthread_mutex_lock(&msmbitem);
    result = mkitem(array, path, SMBITEM_GROUP, "");
    pthread_mutex_unlock(&msmbitem);
    return result;
}

int mkhost(char *path, int flag){
    struct smbitemarray *array = NULL;
    int			result;

    DPRINT("path=%s, flag=%d\n", path, flag);
    path = (char*) GetFirstPath(path, &result);
    if (FindSlash(path)) return 0;
    if (strcmp(path, "") == 0) return 0;
    if (strcmp(path, ".") == 0) return 0;
    if (strcmp(path, "..") == 0) return 0;
    if (flag & AUTOPATH) array = &auto_tree;
    if (flag & USERPATH) array = &user_tree;
    if (array == NULL) return 0;    

    pthread_mutex_lock(&msmbitem);
    result = mkitem(array, path, SMBITEM_COMP, "");
    pthread_mutex_unlock(&msmbitem);
    return result;
    
}

int mklink(char *path, char *link, int flag){
    struct smbitemarray	*array = NULL;
    int			tmp, path_len, result = 0;
    char		*name;
    
    DPRINT("path=%s, link=%s, flag=%d\n", path, link, flag);
    path = (char*) GetFirstPath(path, &path_len);
    name = (char*) GetSecondPath(path, &tmp);
    if (GetPathComponent(path) == 1){
	name = path; path = ""; path_len = 0;
    }

    if (FindSlash(name)) return 0;
    if (strcmp(name, "") == 0) return 0;
    if (strcmp(name, ".") == 0) return 0;
    if (strcmp(name, "..") == 0) return 0;
    if (flag & AUTOPATH) array = &auto_tree;
    if (flag & USERPATH) array = &user_tree;
    if (array == NULL) return 0;    
    
    pthread_mutex_lock(&msmbitem);
    if ((array = finddirlen(array, path, path_len)) != NULL){
	result = mkitem(array, name, SMBITEM_LINK, link);
    }
    pthread_mutex_unlock(&msmbitem);
    return result;
}

void ReleaseSmbItemArray(struct smbitemarray *array){
    if (array == NULL) return;
    pthread_mutex_lock(&msmbitem);
    DestroySmbItemArray(array);
    pthread_mutex_unlock(&msmbitem);
    free(array);
}

struct smbitemarray* getdir(char *path, int flag){
    struct smbitemarray	*result, *array;
    int			error = 1;
    
    if (FindSlash(path)) return NULL;
    if ((result = malloc(sizeof(struct smbitemarray))) == NULL) return NULL;
    memset(result, 0, sizeof(struct smbitemarray));

    pthread_mutex_lock(&msmbitem);
    if (flag & AUTOPATH){
	array = &auto_tree;
	if ((array = finddir(array, path)) != NULL){
	    error = 0;
	    getarray(array, result);
	}
    }	
    if (flag & USERPATH){
	array = &user_tree;
	if ((array = finddir(array, path)) != NULL){
	    error = 0;
	    getarray(array, result);
	}
    }
    pthread_mutex_unlock(&msmbitem);
    if (error){
	free(result);
	result = NULL;
    }
    return result;
}

void ReleaseOldItem(time_t ctime, int flag){
    pthread_mutex_lock(&msmbitem);
    if (flag & AUTOPATH) killold(&auto_tree, ctime);
    if (flag & USERPATH) killold(&user_tree, ctime);
    pthread_mutex_unlock(&msmbitem);
}


int GetPathType(const char *path, void **data){
    struct smbitemarray *array;
    struct smbitem	*item;
    const char		*name;
    int			name_len;
    int			count,i;

    if (data != NULL) *data = NULL;
    count = GetPathComponent(path);
    if (count == 0){
	if (data != NULL) *data = getdir("", AUTOPATH | USERPATH);
	return PATH_TO_DIR;
    }

    name = GetFirstPath(path, &name_len);
    array = getdir("", AUTOPATH | USERPATH);
    item = NULL;
    if (array != NULL){
	if ((i = FindSmbItemLen(array, name, name_len)) >= 0){
	    RequestSmbItem(array->elem[i]);
	    item = array->elem[i];
	};
	ReleaseSmbItemArray(array);
    };
    
    if (item == NULL){
	switch(count){
	    case 1:  return PATH_TO_COMP;
	    case 2:  return PATH_TO_SHARE; 
	    default: return PATH_TO_SMBFILE;
	}
    }

    if (item->type == SMBITEM_COMP){
	ReleaseSmbItem(item);
        switch(count){
	    case 1:  return PATH_TO_COMP;
	    case 2:  return PATH_TO_SHARE;
	    default: return PATH_TO_SMBFILE;
	}
    }

    if (item->type == SMBITEM_LINK){
	if (count != 1){
	    ReleaseSmbItem(item);
	    return PATH_UNKNOWN;
	}
	if (data != NULL){
	    if ((*data = malloc(strlen(item->data) + 1)) != NULL)
		strcpy((char*) *data, item->data);
	};
	ReleaseSmbItem(item);
	return PATH_TO_LINK;
    }

    if (item->type == SMBITEM_GROUP){
	if (count == 3){
	    ReleaseSmbItem(item); 
	    return PATH_UNKNOWN;
	}
	if (count == 1){
	    if (data != NULL)
		*data = getdir((char*)item->name, AUTOPATH | USERPATH);
	    ReleaseSmbItem(item);
	    return PATH_TO_DIR;
	}
	name = GetSecondPath(path, &name_len);
	array = getdir((char*)item->name, AUTOPATH | USERPATH);
	ReleaseSmbItem(item);
	item = NULL;
	if (array != NULL){
	    if ((i = FindSmbItemLen(array, name, name_len)) >= 0){
		RequestSmbItem(array->elem[i]);
		item = array->elem[i];
	    };
	    ReleaseSmbItemArray(array);
	};

	if (item == NULL) return PATH_UNKNOWN;
	if (item->type != SMBITEM_LINK) return PATH_UNKNOWN;
	
	if (data != NULL){
	    if ((*data = malloc(strlen(item->data) + 1)) != NULL)
		strcpy((char*) *data, item->data);
	};
	ReleaseSmbItem(item);
	return PATH_TO_LINK;
    };

    return PATH_UNKNOWN;
}

